/* nm-vpnc-service - vpnc integration with NetworkManager
 *
 * Dan Williams <dcbw@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * (C) Copyright 2005 Red Hat, Inc.
 */

#ifndef NM_VPNC_SERVICE_H
#define NM_VPNC_SERVICE_H

#define NM_DBUS_SERVICE_VPNC	"org.freedesktop.NetworkManager.vpnc"
#define NM_DBUS_INTERFACE_VPNC	"org.freedesktop.NetworkManager.vpnc"
#define NM_DBUS_PATH_VPNC	"/org/freedesktop/NetworkManager/vpnc"

#endif
